/* $Id: nws_cache.c,v 1.3 2002/03/12 15:53:35 mquinson Exp $ */

/* nws_cache - generic data container with multiple indexes */

#include "config.h"
#include <stdio.h> //FIXME
#include "nws_cache_internal.h"
#include <string.h> /* memmove */

NWS_LOG_DECLARE(cache,"");

/*****[ Cell handeling ]******************************************************/
/*----[ actual implementation of a cell ]------------------------------------*/
/* For data container use only. Client code should ignore about cells.       */
/*----[ Access functions ]---------------------------------------------------*/
int nwsDcCellUse(nwsDcCell_t *cell) { /* to tell "I use it" */
  if (!cell)
    return NWSERR_NOTFOUND;
  cell->refcount ++;
  return NWSERR_SUCCES;
}

int nwsDcCellDispose(nwsDcCell_t *cell) {/* to tell "I don't use it anymore"*/
  NWS_VERBOSE(cache)("nwsDcCellDispose(%p)->%d",
		     cell,cell ? cell->refcount-1 : -200 );
  if (!cell)
    return NWSERR_NOTFOUND;
  if (!(--(cell->refcount))) {
    /* Not used anymore, frees it */
    if (cell->data) cell->freeData(cell->data);
    free(cell);
  }
  return NWSERR_SUCCES;
}
      
/*****[ Generic datacontainer ]***********************************************/

/*----[ nwsDcFunction_t ]----------------------------------------------------*/
/* type of interface functions a type have to implement.                     */
/*---------------------------------------------------------------------------*/
typedef struct {
  /* mem management functions (cell handeling by generic type) */
  int (*addCell)(nwsDc_t*self,nwsDcKey_t key,nwsDcCell_t* elem);/* add stuff */
  int (*remove)(nwsDc_t*self,nwsDcCursor_t *cursor);

  int (*retrieveCursor)(nwsDc_t*self,nwsDcKey_t key,
			/*OUT*/nwsDcCursor_t*where);

  /* mem management for all containers */                                    
  int (*commit)(nwsDc_t*self); /* clean up what can be */
  int (*free)(nwsDc_t*self);  /* destructor */

  /* cursor management, for traversal */
  int (*cursorInit)(nwsDc_t*self,/*OUT*/nwsDcCursor_t *cursor);
  int (*cursorNext)(/*IN&OUT*/nwsDcCursor_t *cursor,
		    /*OUT*/nwsDcKey_t *key,nwsDcCell_t **cell);
  int (*cursorFree)(nwsDcCursor_t *cursor);

  /* debugging purpose. if dumpData is null, printf("%s") is used */
  int (*dump)(nwsDc_t*self,void (*dumpData)(void*data));
} nwsDcFunction_t;

/*----[ functions ]----------------------------------------------------------*/
/* Static table to the interface of all possible DC types                    */
/*---------------------------------------------------------------------------*/
static nwsDcFunction_t functions[nwsDcMagicCount] = {
  /* Array type */
  {&nwsDcArrayAdd,&nwsDcArrayRemove,&nwsDcArrayRetrieve,
   &nwsDcArrayCommit,&nwsDcArrayFree,
   &nwsDcArrayCursorInit,&nwsDcArrayCursorNext,&nwsDcArrayCursorFree,
   &nwsDcArrayDump},

  /* Linked list type */
  {&nwsDcListAdd,&nwsDcListRemove,&nwsDcListRetrieve,
   &nwsDcListCommit,&nwsDcListFree,
   &nwsDcListCursorInit,&nwsDcListCursorNext,&nwsDcListCursorFree,
   &nwsDcListDump}


  /* Other types come there */
};

/*----[ internal functions ]-------------------------------------------------*/
/* makes sure the add is done to all copyTo, also                            */
int nwsDcAddCell(nwsDc_t*self,nwsDcKey_t key,nwsDcCell_t* elem);

/*----[ client interface ]---------------------------------------------------*/
/* Interface offered to user code. They mask the differencies between DC,    */
/*  and manage the cells so that the DC don't have to mess with it.          */
/*---------------------------------------------------------------------------*/

int nwsDcAdd(nwsDc_t*self,nwsDcKey_t key,void *data,void (*freeData)(void*)) {
  nwsDcCell_t *cell;

  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }

  /* Malloc the cell */
  cell=malloc(sizeof(nwsDcCell_t));
  if (!cell) 
    NWS_OUT_OF_MEM(cache);
  cell->data=data;
  cell->freeData=freeData;
  cell->refcount=0; /* the guy who asked for this cell will add it */
  cell->valid=1;

  return nwsDcAddCell(self,key,cell);
}

int nwsDcRemove(nwsDc_t*self,nwsDcKey_t key) {
  nwsDcCursor_t cursor;
  int retcode,i;
  
  NWS_VERBOSE(cache)("nwsDcRemove(%p,%s)",self,nwsDcKeyDump(key));
  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }
  
  retcode=functions[self->magic].retrieveCursor(self,key,&cursor);
  if (retcode == NWSERR_NOTFOUND)
    return NWSERR_SUCCES;
  if (retcode != NWSERR_SUCCES)
    return retcode;
  switch (cursor.magic) {
  case nwsDcArrayMagic:
    NWS_VERBOSE(cache)("nwsDcRemove: cursor=%s",
		       nwsDcKeyDump(cursor.owner->t.a.data[cursor.t.a].key));
    break;

  case nwsDcListMagic:
    NWS_VERBOSE(cache)("nwsDcRemove: cursor=%s",
		       nwsDcKeyDump(cursor.t.l.this->key));
  default:
  }

  retcode = functions[self->magic].remove(self,&cursor);
  if (retcode != NWSERR_SUCCES)
    return retcode;

  for (i=0; i<self->copyToCount ;i++) {
    retcode=nwsDcRemove(self->copyTo[i],key);
    if (retcode)
      return retcode;
  }
  if ((retcode = nwsDcCellDispose(cursor.cell))) return retcode;
  nwsDcDump(self,NULL);
  return NWSERR_SUCCES;
}

int nwsDcRetrieve(nwsDc_t*self,nwsDcKey_t key,/*OUT*/void**where) {
  nwsDcCursor_t cursor;
  int retcode;

  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    *where=NULL;
    return NWSERR_BADMAGIC;
  }

  retcode=functions[self->magic].retrieveCursor(self,key,&cursor);
  if (retcode != NWSERR_SUCCES) {
    *where=NULL;
    return retcode;
  }
  if (!cursor.cell->valid) {
    *where=NULL;
    return NWSERR_NOTFOUND;
  }
  
  *where=cursor.cell->data;
  return NWSERR_SUCCES;
}

int nwsDcCopy(nwsDc_t*dest,nwsDcKey_t destKey,nwsDc_t*src,nwsDcKey_t srcKey) {
  nwsDcCursor_t cursor;
  int retcode;

  if (dest->magic < 0 || dest->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",dest->magic);
    return NWSERR_BADMAGIC;
  }
  if (src->magic < 0 || src->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",src->magic);
    return NWSERR_BADMAGIC;
  }

  retcode=functions[src->magic].retrieveCursor(src,srcKey,&cursor);
  if (retcode != NWSERR_SUCCES)
    return retcode;
  if (!cursor.cell->valid)
    return NWSERR_NOTFOUND;

  return nwsDcAddCell(dest,destKey,cursor.cell);
}

int nwsDcCommit(nwsDc_t*self) {
  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }
  //FIXME: traverse(if !valid,remove+dispose)
  return functions[self->magic].commit(self);
}

int nwsDcFree(nwsDc_t *self) { /* Destructor */
  nwsDcCursor_t *c;
  nwsDcKey_t key;
  nwsDcCell_t *cell;
  nwsDc_t *dc;
  int retcode,i;

  NWS_VERBOSE(cache)("nwsDcFree(%p)",self);
  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }
  
  /* Tells the guy I copy to forget me */
  if ((dc = self->copyOf)) {
    for (i=0 ; 
	 i < dc->copyToCount &&
	   dc->copyTo[i] != self ; 
	 i++);
    if (i == dc->copyToCount) {
      NWS_WARN(cache)("Damnit! Can't find myself in self->copyOf->copyTo");
      return NWSERR_NOTFOUND;
    }
    if (i < dc->copyToCount - 1) {
      memmove(dc->copyTo[i], dc->copyTo[i+1],
	      sizeof(nwsDc_t *) * (dc->copyToCount - i - 1));
    }
    dc->copyToCount--;    
    dc->copyTo=realloc(dc->copyTo, dc->copyToCount*sizeof(nwsDc_t *));
  }

  /* traverse(forget me) on copyTo 
     (ie, tells the ones copying me that I'm gone) */
  if (self->copyTo) {
    for (i=0 ; i<self->copyToCount ; i++) {
      self->copyTo[i]->copyOf=NULL;
    }
    free(self->copyTo);
  }
  self->copyToCount=0;

  /* traverse(dispose) on cells */
  if ((retcode=nwsDcCursorNew(self,&c))) {
    NWS_WARN(cache)("Damnit! Can't create a cursor");
    return retcode;
  }
  while (!(retcode=functions[self->magic].cursorNext(c,&key,&cell))) {
    if ((retcode=nwsDcCellDispose(cell)))
      return retcode;
  }
  if (retcode != NWSERR_NOTFOUND) {
    NWS_VERBOSE(cache)("Got err %d in cursorNext",retcode);
    nwsDcCursorFree(c);
    return retcode;
  }
  if ((retcode = nwsDcCursorFree(c))) return retcode;

  /* free the specific parts */
  if ((retcode = functions[self->magic].free(self)) != NWSERR_SUCCES) 
    return retcode;
  /* free the global parts */
  free(self);
  
  return NWSERR_SUCCES;
}

int nwsDcDump(nwsDc_t *self,void (*dumpData)(void*data)) {
  int retcode,i;

  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }
  NWS_BACKTRACE(cache,LogVerbose);
  retcode = functions[self->magic].dump(self,dumpData);
  if (retcode != NWSERR_SUCCES) return retcode;
  if (self->copyOf)
    printf("  copy of %p\n",self->copyOf);
  else 
    printf("  Not a copy\n");
  if (self->copyToCount) {
    printf("List %d copies:\n",self->copyToCount);
    for (i=0 ; i<self->copyToCount ; i++) {
      printf(" >>>  Copy to %p\n",self->copyTo[i]);
      if ((retcode = nwsDcDump(self->copyTo[i],dumpData)))
	return retcode;
      printf(" <<<  End of %p to which %p is copied \n",self->copyTo[i],self);
    }
  } else {
    printf("There is no copy\n");
  }
  return NWSERR_SUCCES;
}

/*----[ for traversal ]------------------------------------------------------*/
int nwsDcCursorNew(nwsDc_t*self,/*OUT*/nwsDcCursor_t **where) {
  if (self->magic < 0 || self->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",self->magic);
    return NWSERR_BADMAGIC;
  }
  *where=malloc(sizeof(nwsDcCursor_t));
  if (!where)
    NWS_OUT_OF_MEM(cache);
  (*where)->magic=self->magic;
  (*where)->owner=self;
  (*where)->cell=NULL;

  return functions[self->magic].cursorInit(self,*where);
}
int nwsDcCursorNext(/*IN&OUT*/nwsDcCursor_t *cursor,
		    /*OUT*/nwsDcKey_t *key,void **data) {
  nwsDcCell_t*cell;
  int retcode;

  if (cursor->magic < 0 || cursor->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",cursor->magic);
    return NWSERR_BADMAGIC;
  }
  if ((retcode=functions[cursor->magic].cursorNext(cursor,key,&cell)))
    return retcode;
  *data=cell->data;
  return NWSERR_SUCCES;
}
int nwsDcCursorFree(nwsDcCursor_t *cursor) {
  int retcode;

  if (cursor->magic < 0 || cursor->magic >= nwsDcMagicCount) {
    NWS_ERROR(cache)("Damnit! Magic number %d not valid.",cursor->magic);
    return NWSERR_BADMAGIC;
  }
  if ((retcode = functions[cursor->magic].cursorFree(cursor)))
    return retcode;
  free(cursor);
  return NWSERR_SUCCES;
}

/*----[ DC interface ]-------------------------------------------------------*/
/* Some generic function all DC need (the key management is in nws_dc_key.c) */
/*---------------------------------------------------------------------------*/
/*....[ nwsDcNew ]...........................................................*/
/* Initialize the fields present in every DC, and take care of copyOf.       */
/*...........................................................................*/
int nwsDcNew(nwsDc_t *dc, 
	     nwsDc_t *copyOf, int (*cmpKey)(nwsDcKey_t,nwsDcKey_t)) {
  
  dc->copyOf = copyOf;
  dc->cmpKey = (cmpKey==NULL ? &nwsDcKeyCmp : cmpKey);
  dc->copyTo = NULL;
  dc->copyToCount = 0;
  if (copyOf) {
    /* register as copyTo to this dc */
    if (copyOf->magic < 0 || copyOf->magic >= nwsDcMagicCount) {
      NWS_ERROR(cache)("Damnit! Magic number %d not valid.",copyOf->magic);
      return NWSERR_BADMAGIC;
    }
    copyOf->copyToCount++;
    copyOf->copyTo=REALLOC(copyOf->copyTo,
			   sizeof(nwsDc_t *)*copyOf->copyToCount);
    if (!copyOf->copyTo)
      NWS_OUT_OF_MEM(cache);
    copyOf->copyTo[copyOf->copyToCount-1]=dc;
  }
  return NWSERR_SUCCES;
}
/*.....[ nwsDcAddCell ]......................................................*/
/* makes sure the add is done to all copyTo, also                            */
/*...........................................................................*/
int nwsDcAddCell(nwsDc_t*self,nwsDcKey_t key,nwsDcCell_t* cell) {
  int i,retcode;

  for (i=0; i<self->copyToCount ;i++) {
    retcode=nwsDcAddCell(self->copyTo[i],key,cell);
    if (retcode) return retcode;
  }
  if ((retcode=nwsDcCellUse(cell))) return retcode;
  return functions[self->magic].addCell(self,key,cell);
}

