/* $Id: nws_cache_internal.h,v 1.3 2002/03/12 15:53:35 mquinson Exp $ */

/* nws_cache - generic data container with multiple indexes                  */

/* this file contain stuff that every datacontainer type must see, but not   */
/*   the client code                                                         */

#ifndef __NWS_CACHE_INTERNAL_H
#define __NWS_CACHE_INTERNAL_H
#include "nws_cache.h" /* datacontainer can access the public interface */

#include "nws_dc_array.h" /* to load the interface to array */
#include "nws_dc_list.h"  /* to load the interface to list */

/*****[ key handeling ]*******************************************************/
/** All keys are of type nwsDcKey_t, described in public interface          **/
/*****************************************************************************/

/*----[ nwsDcKeyCmp ]--------------------------------------------------------*/
/* The good old strcmp like function.                                        */
/* Numeric values are compared using < and >.                                */
/* String ones are compared using strcmp.                                    */
/* Void* can't be compared (return -2 in any case).                          */
/*                                                                           */
/* Return value:                                                             */
/*    1: a >  b                                                              */
/*    0: a == b                                                              */
/*   -1: a <  b                                                              */
/*   -2: a and b can't be compared.                                          */
/*       If you're sure of what you're comparing, it will never happen.      */
/*---------------------------------------------------------------------------*/
extern int nwsDcKeyCmp(nwsDcKey_t a,nwsDcKey_t b);

/*----[ nwsDcKeyDump ]-------------------------------------------------------*/
/* Usefull only when type != void*. Result is overwritten from call to call  */
/*---------------------------------------------------------------------------*/
char *nwsDcKeyDump(nwsDcKey_t k);

/*****[ magic values ]********************************************************/
/* Since we are casting the types around, let's get secure and               */
/*  put a magic indentificator at the begining of the structures             */
/*****************************************************************************/
typedef enum {
  /* container_plain stuff */
  nwsDcArrayMagic = 0,  /* Flat array */
  nwsDcListMagic = 1,  /* Simple linked list */

  nwsDcMagicCount=2 /* sentinel */
} nwsDcMagic_t;

/*****[ Cell handeling ]******************************************************/
/** All cell are of type nwsDcCell_t (fields readonly when not in cache.c)  **/
/** The management of this stuff is done by the generic data container.     **/
/**  They should never be written directly by datacontainer implementations **/
/**  The user don't even need to know about them.                           **/
/*****************************************************************************/
struct nwsDcCell_ {
  void *data;
  void (*freeData)(void*);
  int refcount;
  int valid;
};

/*----[ Access functions ]---------------------------------------------------*/
int nwsDcCellNew(nwsDcKey_t key,void *data,void (*freeData)(void*),
		 /* OUT */ nwsDcCell_t **where);

int nwsDcCellUse(nwsDcCell_t *cell); /* to tell "I use it" */
int nwsDcCellDispose(nwsDcCell_t *cell); /* to tell "I don't use it anymore"*/

/*****[ Generic datacontainer ]***********************************************/
struct nwsDc_ {
  nwsDcMagic_t magic; 
  nwsDc_t *copyOf;    
  nwsDc_t **copyTo;   
  int copyToCount;    
  int (*cmpKey)(nwsDcKey_t,nwsDcKey_t);
  union {
    nwsDcArray_ctn a;
    nwsDcList_ctn l;
  } t; /* type */
};

/*----[ nwsDcNew ]-----------------------------------------------------------*/
/* Initialize the fields present in every DC, and take care of copyOf.       */
/*---------------------------------------------------------------------------*/
int nwsDcNew(nwsDc_t *dc, 
	     nwsDc_t *copyOf, int (*cmpKey)(nwsDcKey_t,nwsDcKey_t));


/*****[ Cursor ]**************************************************************/
/* access functions are part of the public interface */

/* Virtual datatype, casted around */
struct nwsDcCursor_ {
  nwsDcMagic_t magic;
  nwsDc_t *owner;
  nwsDcCell_t *cell;
  union {
    nwsDcCursorArray_ctn a;
    nwsDcCursorList_ctn l;
  } t; /* type */
};

#endif /* __NWS_CACHE_INTERNAL_H */
