/* $Id: nws_log.c,v 1.5 2002/03/12 15:24:00 mquinson Exp $ */

/* log - logging utility functions                                           */

/* License: dunno, ask Rich.                                                 */

#include <stdio.h>
#include <time.h>      /* time() */
#include <unistd.h>    /* getpid() */
#include <stdarg.h>    /* va_*, vfprintf() */
#include <execinfo.h>  /* to print the backtrace */

#include "nws_err.h" /* nws errors utils */
#include "strutil.h" /* nws string utils */
#include "nws_log.h" /* nws log utils    */
#include "nws_log_channels.h" /* generated automatically lists the channels */

NWS_LOG_DECLARE(log,"");
__NWS_LOG_DO_DECLARES;  /* declare the __nws_lchan_##name (they live here and
			   are external in all other locations) */

/*----[ Level declarations ]-------------------------------------------------*/
/* What to print when displayed */
const char *nws_log_level_prefix[LogLevelCount] =
  {"verbose","debug","WARNING","ERROR"};
/* To lookup level from its name */
const char *nws_log_level_name[LogLevelCount] = {"verbose","debug","warn","error"};


#define MAX_LOG_REGISTER 30
static nws_log_channel_t* channels_registered[MAX_LOG_REGISTER];

/* The position is globalized, so that the NWS_LOG macro can rewrite itself
   to a function having exactly the printf prototype */
nws_log_channel_t _nws_log_channel; /* on which channel we write */
int _nws_log_level;  /* at which level we log */
const char *_nws_log_file; /* position of the logging macro */
int _nws_log_line;  /* idem */

/*----[ nws_log_printf ]-----------------------------------------------------*/
void nws_log_printf(const char *fmt,...) {
  va_list args;

  fprintf(_nws_log_channel.stream[_nws_log_level],
	  "%d:%s:%d %s/%s ",(int)time(NULL),_nws_log_file,_nws_log_line,
	 _nws_log_channel.name,nws_log_level_name[_nws_log_level]);
  va_start(args,fmt);
  vfprintf(_nws_log_channel.stream[_nws_log_level],fmt,args);
  va_end(args);
  fprintf(_nws_log_channel.stream[_nws_log_level],"\n");
  fflush(_nws_log_channel.stream[_nws_log_level]);
}

/* Parse the options and apply the changes to the channel corresponding
** if #channel# is null, it will be founded from the table
** If the options are unparsable, emmit a warning on (log), and return 
** NWSERR_PARSE_ERROR.
*/
int __nws_log_parse_option(nws_log_channel_t *channel,const char *optionSet) {
  int levelnum; /* numerical value of level */
  int found,i;
  nws_log_channel_t *lastchan=NULL; /* last channel seen. The one which will be changed */ 
  char *buff=strdup(optionSet);
  char *option;
  FILE *valstream;

  NWS_VERBOSE(log)("nws_log_parse_option(%s,\"%s\")",channel?channel->name:"NULL",optionSet);

  if (!buff)  NWS_OUT_OF_MEM(log);
  if (channel) lastchan=channel;
  
  for(option = strtok(buff, ",");
      option != NULL;
      option = strtok(NULL, ",")) {
    char *level = strchr(option,':');
    char *value = strchr(level ? level : option, '=');
    if (!value) {
      NWS_ERROR(log)("Parse error in logging option \"%s\". No value specified.",option);
      return NWSERR_PARSE;
    }
    if (level) {
      *level++='\0';
    } else {
      level=option;
      option=NULL;
    }

    NWS_VERBOSE(log)("option=%s",option);
    if (!option && lastchan) {
	option=(char *)lastchan->name;
    } else if (!option && !lastchan) {
      NWS_ERROR(log)("Parse error in logging option \"%s\". No channel specified.",optionSet);
      return NWSERR_PARSE;
    } else { /* option is set, change lastchan */
      found=0;
      for (i=0 ; 
	   i<MAX_LOG_REGISTER && channels_registered[i];
	   i++) {
	NWS_VERBOSE(log)("#%d: '%s' ?= '%s'",i,channels_registered[i]->name,option);
	if (strmatch(channels_registered[i]->name,option)) {
	  if (found) {
	    NWS_ERROR(log)
	      ("Parse error in logging option \"%s\". Can't change several channels at the same time.",optionSet);
	    return NWSERR_PARSE;
	  }
	  lastchan=channels_registered[i];
	  found=1;
	}
      }
      if (!found) {
	NWS_ERROR(log)("Parse error in logging options \"%s\". Inexistant logging channel.",optionSet);
	return NWSERR_PARSE;
      }
    }

    *value++='\0'; /* Separate level and value by a \0, and put value to the right point */
    valstream=(FILE*)-1;
    if (!strcmp(value,"2") || !strcmp(value,"stderr")) valstream=stderr;
    if (!strcmp(value,"1") || !strcmp(value,"stdout")) valstream=stdout;
    if (!strcmp(value,"0") || !strcmp(value,"null") ) valstream=NULL;
    //printf("Value=%s;func=%p\n",value,valfunc);
    //    printf("null=%p;stdout=%p;stderr=%p\n",&printf_null,&printf_stdout,&printf_stderr);
    if (valstream==(FILE*)-1) {
      NWS_ERROR(log)("Parse error in logging option \"%s\". "
		     "Value must be stderr,stdout or null.",optionSet);
      return NWSERR_PARSE;
    }


    if (!strcmp(level,"")) {
      NWS_ERROR(log)("Parse error in logging option \"%s\". No level specified.",optionSet);
      return NWSERR_PARSE;
    }

    found=0;
    for (levelnum=0 ; levelnum<LogLevelCount ; levelnum++) {
      if (strnmatch(nws_log_level_name[levelnum],level,value-level-1)) {
	NWS_VERBOSE(log)("Change level %s of %s to %s",
			 nws_log_level_name[levelnum],(option?option:lastchan->name),value);
	lastchan->stream[levelnum]=valstream;
	found=1;
      }
    }
    if (!found) {
      NWS_ERROR(log)("Parse error in logging option \"%s\". Unknown level.",optionSet);
      return NWSERR_PARSE;
    }
  }

  NWS_VERBOSE(log)("---------");
  free(buff);
  return NWSERR_SUCCES;
}

/* Add a channel to the main table of channels, and apply #options# to it.
** Returns what nws_log_parse_option does. 
*/
int nws_log_register(const char *name,const char *options,
			     nws_log_channel_t *channel) {
  int i;
  
  /* search a free slot or the channel already registred */
  for (i=0;
       i<MAX_LOG_REGISTER && channels_registered[i] && channels_registered[i]->name && 
	 strcmp(channels_registered[i]->name,name);
       i++);
  if (i == MAX_LOG_REGISTER) {
    NWS_ERROR(log)("Too much channels. Current maximum is %d."
		    " You can increase it by changing the MAX_LOG_REGISTER value in %s",
		    MAX_LOG_REGISTER,__FILE__);
  }
  channels_registered[i]=channel;

  /* Apply defaults */
  channel->stream[LogVerbose]=NULL;
  channel->stream[LogDebug]=NULL;
  channel->stream[LogWarn]=stdout;
  channel->stream[LogError]=stderr;

  return __nws_log_parse_option(channel,options);
}

/* Parse the options concerning the logging sub-system */
void nws_log_init(int* argc_addr,const char **argv) {
    int i;
    /* cleanup */
    for (i=0; i<MAX_LOG_REGISTER; i++)
      channels_registered[i]=NULL;

    /* load the defined channel */
    __NWS_LOG_REGISTERS;

    /* parse the options */
    for (i=0 ; i<*argc_addr ;i++) {
      if (strncmp(argv[i],"--log=",strlen("--log="))) 
	continue;

      /* found */
      __nws_log_parse_option(NULL,argv[i] + strlen("--log="));
    }

    /* dump it */
    if (NWS_VERBOSE_ON(log)) {
      for (i=0; i<MAX_LOG_REGISTER; i++) 
	if (channels_registered[i]) {
	  printf("channels_registered[%d]->name=%s\n",i,channels_registered[i]->name);
	  printf("channels_registered[%d]->verbose=%p\n",
		 i,channels_registered[i]->stream[LogVerbose]);
	  printf("channels_registered[%d]->debug=%p\n",
		 i,channels_registered[i]->stream[LogDebug]);
	  printf("channels_registered[%d]->warn=%p\n",
		 i,channels_registered[i]->stream[LogWarn]);
	  printf("channels_registered[%d]->error=%p\n\n",
		 i,channels_registered[i]->stream[LogError]);
	}
      printf("FIXME: better names for channels in previous dumping\n");
    }
}

void __nws_backtrace(nws_log_channel_t channel) {
 void *array[30];
  size_t size;
  char **strings;
  size_t i;

  size = backtrace (array, 30);
  strings = backtrace_symbols (array, size);

  printf("----[ Begin Backtrace ]----\n");
  for (i = 0; i < size; i++)
     printf ("%s\n", strings[i]);
  printf("----[ End Backtrace ]----\n");

  free (strings);
}

